/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* misc.cpp */

/* Includes little-used or obsolete routines */

/* Includes */
#include "struct.h"

void Distribution::UpdateEmissionNestedDivide( Data* data, double** prob_s, double norm_const ) {

  double**** pair_prob;         // Projection on the pair of the variables
  double** MI;                  // Mutual information
  long** sorted_MI;             // Indices of the sorted mutual information entries
  long** connected;             // Connectivity matrix
  long max_num_edges;           // Maximum number of edges allowed for Chow-Liu distribution
  long* best_edge;              // The list of current best edges

  double*** mix_comp_prob;
  
  long *group, *num_nodes, **node_list;
  double baseline_ll;
  
  /* Index variable(s) */
  long i,j,n,t,b,i1,i2;
  
  /* Temporary variable(s) */
  double sum, best_MI;

  /* External variables */
  extern double cl_threshold;   // !!! Mutual information threshold for Chow-Liu trees !!!

  /* Deallocating the current structure -- will be rebuilt from scratch */
  for( i=0; i<num_dists; i++ )
    delete subdist[i];
  delete subdist;

  /* Building a Chow-Liu tree */
  /* Allocating individual probabilities */
  ind_prob=new double*[dim];
  for( i=0; i<dim; i++ )
    ind_prob[i]=new double[num_states];

  /* Allocating the array of pairwise probabilities */
  pair_prob=new double***[dim];
  for( i=0; i<dim; i++ )
    {
      pair_prob[i]=new double**[i];
      for( j=0; j<i; j++ )
	{
	  pair_prob[i][j]=new double*[num_states];
	  for( i1=0; i1<num_states; i1++ )
	    pair_prob[i][j][i1]=new double[num_states];
	}
    }
  
  /* Allocating mutual information lower triangular matrix */
  MI=new double*[dim];
  for( i=0; i<dim; i++ )
    MI[i]=new double[dim];
    
  edge=new long*[dim-1];
  for( i=0; i<dim-1; i++ )
    edge[i]=new long[2];
  edge_MI=new double[dim-1];
  
  /* Array of best edges */
  best_edge=new long[dim-1];
    
  /* Initializing */
  
  /* Individual probability projections */
  for( i=0; i<dim; i++ )
    for( i1=0; i1<num_states; i1++ )
      ind_prob[i][i1]=0.0;
  
  /* Pairwise probability projections */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      for( i1=0; i1<num_states; i1++ )
	for( i2=0; i2<num_states; i2++ )
	  pair_prob[i][j][i1][i2]=0.0;
  
  /* Mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      MI[i][j]=0.0;
  
  /* Looping through the data to collect the counts */
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      for( i=0; i<dim; i++ )
	{ /* !!! Checking for missing data !!! */
	  if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[i]] ) )
	    {
	      ind_prob[i][data->sequence[n]->entry[t]->ddata[dim_index[i]]]+=prob_s[n][t];
	      for( j=0; j<i; j++ )
		/* !!! !!! */
		if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[j]] ) )
		  pair_prob[i][j][data->sequence[n]->entry[t]->ddata[dim_index[i]]]
		    [data->sequence[n]->entry[t]->ddata[dim_index[j]]]+=prob_s[n][t];
	    }
	} /* !!! Checking for missing data !!! */
  
  /* Calculating probability distribution projections from the counts */
  /* The normalizing factor is the normalizing constant! */
  for( i=0; i<dim; i++ )
    {
      sum=0.0;
      for( i1=0; i1<num_states; i1++ )
	sum+=ind_prob[i][i1];
      
      for( i1=0; i1<num_states; i1++ )
	ind_prob[i][i1]/=sum;
    }
  
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      {
	sum=0.0;
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    sum+=pair_prob[i][j][i1][i2];
	
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    pair_prob[i][j][i1][i2]/=sum;
      }
  
  /* Calculating mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      for( i1=0; i1<num_states; i1++ )
	{
	  MI[i][j]-=xlogx(ind_prob[i][i1]);
	  MI[i][j]-=xlogx(ind_prob[j][i1]);
	  for( i2=0; i2<num_states; i2++ )
	    MI[i][j]+=xlogx(pair_prob[i][j][i1][i2]);
	}
    
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      MI[j][i]=MI[i][j];
  
  node_used=new long[dim];
  
  /* Initializing the list of connected nodes */
  node_used[dim-1]=1;
  for( i=0; i<dim-1; i++ )
    node_used[i]=0;
  
  /* Initializing the array of best edges */
  /* Current best edge is always from the only "attached" node */
  for( i=0; i<dim-1; i++ )
    best_edge[i]=dim-1;
  
  /* Inserting dim-1 edges */
  for( num_edges=0; num_edges<dim-1; num_edges++ )
    {
      best_MI=NEG_INF;
      i1=-1;
      for( i=0; i<dim-1; i++ )
	if( !node_used[i] )
	  if( MI[i][best_edge[i]]>best_MI )
	    { /* Found a better edge */
	      i1=i;
	      best_MI=MI[i][best_edge[i]];
	    } /* Found a better edge */
      
      /* Adding the edge */
      node_used[i1]=1;
      if( i1>best_edge[i1] )
	{
	  edge[num_edges][0]=i1;
	  edge[num_edges][1]=best_edge[i1];
	}
      else
	{
	  edge[num_edges][0]=best_edge[i1];
	  edge[num_edges][1]=i1;
	}
      
      /* Storing the information about the edge */
      edge_MI[num_edges]=best_MI;
      /* Rank is not yet available */
      
      /* Adjusting the list of best edges to not-yet-connected nodes */
      for( i=0; i<dim-1; i++ )
	if( !node_used[i] )
	  if( MI[i][i1]>MI[i][best_edge[i]] )
	    best_edge[i]=i1;
    }
  
  /* Separating a Chow-Liu tree into strong clusters by dropping weak links */
  
  /* Allocating and initializing groups of nodes and node counts */
  num_dists=0;
  group=new long[dim];
  node_list=new long*[dim+1];
  node_list[num_dists]=new long[dim];
  num_nodes=new long[dim];
  for( i=0; i<dim; i++ )
    {
      group[i]=-1;
      num_nodes[i]=0;
    }
  
  for( num_edges=0; num_edges<dim-1; num_edges++ )
    {
      if( edge_MI[num_edges]>DIVIDE_MI_THRESHOLD )
	{ /* Adding nodes to the same group */
	  if( group[edge[num_edges][0]]!=-1 )
	    {
	      group[edge[num_edges][1]]=group[edge[num_edges][0]];
	      node_list[group[edge[num_edges][0]]][num_nodes[group[edge[num_edges][0]]]]=
		edge[num_edges][1];
	      num_nodes[group[edge[num_edges][0]]]++;
	    }
	  else if( group[edge[num_edges][1]]!=-1 )
	    {
	      group[edge[num_edges][0]]=group[edge[num_edges][1]];
	      node_list[group[edge[num_edges][1]]][num_nodes[group[edge[num_edges][1]]]]=
		edge[num_edges][0];
	      num_nodes[group[edge[num_edges][1]]]++;
	    }
	  else
	    { /* Neither node is assigned to a group */
	      group[edge[num_edges][0]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=edge[num_edges][0];
	      num_nodes[num_dists]++;
	      group[edge[num_edges][1]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=edge[num_edges][1];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    } /* Neither node is assigned to a group */
	} /* Adding nodes to the same group */
      else
	{ /* Adding nodes to different groups */
	  if( group[edge[num_edges][0]]!=-1 )
	    {
	      group[edge[num_edges][1]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=edge[num_edges][1];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    }
	  else if( group[edge[num_edges][1]]!=-1 )
	    {
	      group[edge[num_edges][0]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=edge[num_edges][0];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    }
	  else
	    { /* Neither node is assigned to a group */
	      group[edge[num_edges][0]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=edge[num_edges][0];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	      group[edge[num_edges][1]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=edge[num_edges][1];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    } /* Neither node is assigned to a group */
	} /* Adding nodes to different groups */
    }      
  
  /* Deallocating mutual information */
  for( i=0; i<dim; i++ )
    delete MI[i];
  delete MI;
  
  delete node_used;
  
  delete edge_MI;
  
  for( i=0; i<dim-1; i++ )
    delete edge[i];
  delete edge;
  
  /* Deallocating the array of best edges */
  delete best_edge;
  
  /* Deallocating groups counts */
  delete group;
  
  /* Train a model for each cluster */
  
  subdist=new Distribution*[num_dists];
  
  for( i=0; i<num_dists; i++ )
    /* Creating a distribution for each group */
    if( num_nodes[i]==1 )
      { /* Single node -- creating univariate Bernoulli */
	
	subdist[i]=new Distribution( DIST_BERNOULLI, num_states, num_nodes[i] );
	for( j=0; j<num_nodes[i]; j++ )
	  subdist[i]->dim_index[j]=dim_index[node_list[i][j]];
	
	/* Updating the parameters */
	for( j=0; j<num_states; j++ )
	  subdist[i]->state_prob[j]=ind_prob[node_list[i][0]][j];
      } /* Single node -- creating univariate Bernoulli */
    else if( num_nodes[i]==2 || num_nodes[i]==3 || norm_const<=MIN_COMPONENT_SIZE )
      { /* Too few nodes or too few data points for potential splitting -- Chow-Liu tree */
	
	subdist[i]=new Distribution( DIST_FACTOR, 0, num_nodes[i] );
	for( j=0; j<num_nodes[i]; j++ )
	  {	    
	    subdist[i]->subdist[j]=new Distribution( DIST_BERNOULLI, num_states, 1 );
	    subdist[i]->subdist[j]->dim_index[0]=dim_index[node_list[i][j]];
	  }
	
	/* Updating the parameters */
	for( j=0; j<num_nodes[i]; j++ )
	  for( b=0; b<num_states; b++ )
	    subdist[i]->subdist[j]->state_prob[b]=ind_prob[node_list[i][j]][b];
      } /* Too few nodes or too few data points for potential splitting -- Chow-Liu tree */
    else
      { /* Creating a mixture of distributions */
	
	/* First, learning a Chow-Liu tree! */
	/* !!! Can be replaced with the Chow-Liu tree already computed for the component !!! */
	subdist[i]=new Distribution( DIST_CHOWLIU, num_states, num_nodes[i] );
	for( j=0; j<num_nodes[i]; j++ )
	  subdist[i]->dim_index[j]=dim_index[node_list[i][j]];
	
	/* Updating the parameters */
	subdist[i]->UpdateEmissionParameters( data, prob_s, norm_const );
	
	/* Calculating weighted log-likelihood under the Chow-Liu tree */
	baseline_ll=subdist[i]->weighted_ll( data, prob_s );
	
	/* Deallocating the Chow-Liu tree model */
	delete subdist[i];
	
	/* !!! Assuming a mixture of 2 components !!! */
	subdist[i]=new Distribution( DIST_MIXTURE, 2, num_nodes[i] );
	/* Allocating sub-distributions of the mixture -- starting with Chow-Liu trees*/
	for( b=0; b<2; b++ )
	  {
	    subdist[i]->subdist[b]=new Distribution( DIST_CHOWLIU, num_states, num_nodes[i] );
	    for( j=0; j<num_nodes[i]; j++ )
	      subdist[i]->subdist[b]->dim_index[j]=dim_index[node_list[i][j]];
	  }
	
	/* Need to learn parameters at least as good as a Chow-Liu tree! */
	/* Initializing the distribution */
	subdist[i]->Initialize( data );
	
	/* Learning the parameters to be at least as good as a single component Chow-Liu tree */
	subdist[i]->LearnGoodParameters( data, prob_s, norm_const, baseline_ll );
	
	/* Allocating the weights */
	mix_comp_prob=new double**[2];
	for( b=0; b<2; b++ )
	  {
	    mix_comp_prob[b]=new double*[data->num_seqs];
	    for( n=0; n<data->num_seqs; n++ )
	      mix_comp_prob[b][n]=new double[data->sequence[n]->seq_length];
	  }
	
	/* Calculating the weights of the data according to the mixture */
	/* Another E-step */
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    {
	      sum=0.0;
	      for( b=0; b<2; b++ )
		{
		  if( t==0 )
		    mix_comp_prob[b][n][t]=subdist[i]->state_prob[b]*
		      subdist[i]->subdist[b]->prob( data->sequence[n]->entry[t], NULL );
		  else
		    mix_comp_prob[b][n][t]=subdist[i]->state_prob[b]*
		      subdist[i]->subdist[b]->prob( data->sequence[n]->entry[t], 
						    data->sequence[n]->entry[t-1] );
		  sum+=mix_comp_prob[b][n][t];
		}
	      
	      for( b=0; b<2; b++ )
		mix_comp_prob[b][n][t]/=sum;
	    }
	
	/* Passing weights of the data to the mixture components and creating a mixture of potentially
	   divisible components */
	for( b=0; b<2; b++ )
	  {
	    sum=0.0;
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=0; t<data->sequence[n]->seq_length; t++ )
		{	    
		  mix_comp_prob[b][n][t]*=prob_s[n][t];
		  sum+=mix_comp_prob[b][n][t];
		}
	    
	    delete subdist[i]->subdist[b];
	    subdist[i]->subdist[b]=new Distribution( DIST_NESTEDDIVIDE, num_states, num_nodes[i] );
	    for( j=0; j<num_nodes[i]; j++ )
	      subdist[i]->subdist[b]->dim_index[j]=dim_index[node_list[i][j]];
	    
	    /* Updating mixture components */
	    subdist[i]->subdist[b]->UpdateEmissionParameters( data, mix_comp_prob[b], sum );
	  }
	
	/* Deallocating the weights */
	for( b=0; b<2; b++ )
	  {
	    for( n=0; n<data->num_seqs; n++ )
	      delete mix_comp_prob[b][n];
	    delete mix_comp_prob[b];
	  }
	delete mix_comp_prob;
	
      } /* Creating a mixture of distributions */
  
  /* Deallocating individual probabilities */
  for( i=0; i<dim; i++ )
    delete ind_prob[i];
  delete ind_prob;
  
  /* Deallocaing pairwise probabilities */
  for( i=0; i<dim; i++ )
    {
      for( j=0; j<i; j++ )
	{
	  for( i1=0; i1<num_states; i1++ )
	    delete pair_prob[i][j][i1];
	  delete pair_prob[i][j];
	}
      delete pair_prob[i];
    }
  delete pair_prob;
  
  /* Deallocating node counts */
  for( i=0; i<num_dists+1; i++ )
    delete node_list[i];
  delete node_list;
  delete num_nodes;
  
  return;
}

void Distribution::LearnGoodParameters( Data* data, double** prob_s, double norm_const, double base_ll ) {
  /* Learns parameters of the model such that weighted log-likelihood is at least base_ll */

  double weighted_prob, prev_weighted_prob;
  double num_weighted_entries;
  double epsilon;

  //  printf( "Learning the parameters for the mixture to exceed weighted log-likelihood of %le\n", base_ll );
  //  printf( "Number of examples in the component: %le\n", norm_const );
  
  /* Only applicable to mixture distributions */
  if( type==DIST_MIXTURE )
    {
      num_weighted_entries=data->weighted_total_entries( prob_s );
      epsilon=EM_MIX_EPSILON*num_weighted_entries;

      weighted_prob=weighted_ll( data, prob_s );
      prev_weighted_prob=NEG_INF;

      do
	{
	  if( weighted_prob<prev_weighted_prob+epsilon )
	    { /* Initializing the distribution */
	      Initialize( data );
	      weighted_prob=weighted_ll( data, prob_s );
	      prev_weighted_prob=NEG_INF;
	    } /* Initializing the distribution */
	  else
	    {
	      prev_weighted_prob=weighted_prob;
	      UpdateEmissionParameters( data, prob_s, norm_const );
	      weighted_prob=weighted_ll( data, prob_s );
	    }
	}
      while( weighted_prob<base_ll );

#ifdef ALTERNATIVE
      do
	{
	  Initialize( data );
	  weighted_prob=weighted_ll( data, prob_s );
	  do
	    {
	      prev_weighted_prob=weighted_prob;
	      UpdateEmissionParameters( data, prob_s, norm_const );
	      weighted_prob=weighted_ll( data, prob_s );	      
	    }
	  while( weighted_prob>prev_weighted_prob+epsilon );
	}
      while( weighted_prob<base_ll );
#endif
    }
  else
    UpdateEmissionParameters( data, prob_s, norm_const );

  return;
}

void Distribution::UpdateNestedParameters( Data* data, double** prob_s, double norm_const, 
					   Distribution* CLtree ) {

  Distribution* temp_dist;
  long *group, **node_list, *num_nodes;
  double baseline_ll;
  double*** mix_comp_prob;
  double sum;

  /* Index variable(s) */
  long i,j, num_edges, b, n, t;

  /* Deallocating the current structure -- will be rebuilt from scratch */
  for( i=0; i<num_dists; i++ )
    delete subdist[i];
  delete subdist;
  
  if( !CLtree )
    { /* Building a Chow-Liu tree */
      CLtree=new Distribution( DIST_CHOWLIU, num_states, dim );
      for( i=0; i<dim; i++ )
	CLtree->dim_index[i]=dim_index[i];
      CLtree->UpdateEmissionParameters( data, prob_s, norm_const );
    } /* Building a Chow-Liu tree */


  /* Separating a Chow-Liu tree into strong clusters by dropping weak links */

  /* Allocating and initializing groups of nodes and node counts */
  num_dists=0;
  group=new long[dim];
  node_list=new long*[dim+1];
  node_list[num_dists]=new long[dim];
  num_nodes=new long[dim];
  for( i=0; i<dim; i++ )
    {
      group[i]=-1;
      num_nodes[i]=0;
    }
    
  for( num_edges=0; num_edges<dim-1; num_edges++ )
    {
      if( CLtree->edge_MI[num_edges]>DIVIDE_MI_THRESHOLD )
	{ /* Adding nodes to the same group */
	  if( group[CLtree->edge[num_edges][0]]!=-1 )
	    {
	      group[CLtree->edge[num_edges][1]]=group[CLtree->edge[num_edges][0]];
	      node_list[group[CLtree->edge[num_edges][0]]][num_nodes[group[CLtree->edge[num_edges][0]]]]=
		CLtree->edge[num_edges][1];
	      num_nodes[group[CLtree->edge[num_edges][0]]]++;
	    }
	  else if( group[CLtree->edge[num_edges][1]]!=-1 )
	    {
	      group[CLtree->edge[num_edges][0]]=group[CLtree->edge[num_edges][1]];
	      node_list[group[CLtree->edge[num_edges][1]]][num_nodes[group[CLtree->edge[num_edges][1]]]]=
		CLtree->edge[num_edges][0];
	      num_nodes[group[CLtree->edge[num_edges][1]]]++;
	    }
	  else
	    { /* Neither node is assigned to a group */
	      group[CLtree->edge[num_edges][0]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=CLtree->edge[num_edges][0];
	      num_nodes[num_dists]++;
	      group[CLtree->edge[num_edges][1]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=CLtree->edge[num_edges][1];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    } /* Neither node is assigned to a group */
	} /* Adding nodes to the same group */
      else
	{ /* Adding nodes to different groups */
	  if( group[CLtree->edge[num_edges][0]]!=-1 )
	    {
	      group[CLtree->edge[num_edges][1]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=CLtree->edge[num_edges][1];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    }
	  else if( group[CLtree->edge[num_edges][1]]!=-1 )
	    {
	      group[CLtree->edge[num_edges][0]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=CLtree->edge[num_edges][0];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    }
	  else
	    { /* Neither node is assigned to a group */
	      group[CLtree->edge[num_edges][0]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=CLtree->edge[num_edges][0];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	      group[CLtree->edge[num_edges][1]]=num_dists;
	      node_list[num_dists][num_nodes[num_dists]]=CLtree->edge[num_edges][1];
	      num_nodes[num_dists]++;
	      num_dists++;
	      node_list[num_dists]=new long[dim];
	    } /* Neither node is assigned to a group */
	} /* Adding nodes to different groups */
    }      
  
  /* Train a model for each cluster */
  
  subdist=new Distribution*[num_dists];
  
  for( i=0; i<num_dists; i++ )
    /* Creating a distribution for each group */
    if( num_nodes[i]==1 )
      { /* Single node -- creating univariate Bernoulli */
	
	subdist[i]=new Distribution( DIST_BERNOULLI, num_states, num_nodes[i] );
	for( j=0; j<num_nodes[i]; j++ )
	  subdist[i]->dim_index[j]=dim_index[node_list[i][j]];
	
	/* Updating the parameters */
	subdist[i]->UpdateEmissionParameters( data, prob_s, norm_const );
	//	  for( j=0; j<num_states; j++ )
	//  subdist[i]->state_prob[j]=ind_prob[node_list[i][0]][j];
      } /* Single node -- creating univariate Bernoulli */
    else if( num_nodes[i]==2 || num_nodes[i]==3 || norm_const<=MIN_COMPONENT_SIZE )
      { /* Too few nodes or too few data points for potential splitting -- Chow-Liu tree */
	
	subdist[i]=new Distribution( DIST_CHOWLIU, num_states, num_nodes[i] );
	for( j=0; j<num_nodes[i]; j++ )
	  subdist[i]->dim_index[j]=dim_index[node_list[i][j]];
	
	/* Updating the parameters */
	subdist[i]->UpdateEmissionParameters( data, prob_s, norm_const );
      } /* Too few nodes or too few data points for potential splitting -- Chow-Liu tree */
    else
      { /* Creating a mixture of distributions */
	  
	/* First, learning a Chow-Liu tree! */
	/* !!! Can be replaced with the Chow-Liu tree already computed for the component !!! */
	subdist[i]=new Distribution( DIST_CHOWLIU, num_states, num_nodes[i] );
	for( j=0; j<num_nodes[i]; j++ )
	  subdist[i]->dim_index[j]=dim_index[node_list[i][j]];
	
	/* Updating the parameters */
	subdist[i]->UpdateEmissionParameters( data, prob_s, norm_const );
	
	/* Calculating weighted log-likelihood under the Chow-Liu tree */
	baseline_ll=subdist[i]->weighted_ll( data, prob_s );
	
	/* Deallocating the Chow-Liu tree model */
	delete subdist[i];
	
	/* !!! Assuming a mixture of 2 components !!! */
	subdist[i]=new Distribution( DIST_MIXTURE, 2, num_nodes[i] );
	/* Allocating sub-distributions of the mixture -- starting with Chow-Liu trees*/
	for( b=0; b<2; b++ )
	  {
	    subdist[i]->subdist[b]=new Distribution( DIST_CHOWLIU, num_states, num_nodes[i] );
	    for( j=0; j<num_nodes[i]; j++ )
	      subdist[i]->subdist[b]->dim_index[j]=dim_index[node_list[i][j]];
	  }

	/* Need to learn parameters at least as good as a Chow-Liu tree! */
	/* Initializing the distribution */
	subdist[i]->Initialize( data );
	
	/* Learning the parameters to be at least as good as a single component Chow-Liu tree */
	subdist[i]->LearnGoodParameters( data, prob_s, norm_const, baseline_ll );
	
	/* Allocating the weights */
	mix_comp_prob=new double**[2];
	for( b=0; b<2; b++ )
	  {
	    mix_comp_prob[b]=new double*[data->num_seqs];
	    for( n=0; n<data->num_seqs; n++ )
	      mix_comp_prob[b][n]=new double[data->sequence[n]->seq_length];
	  }
	
	/* Calculating the weights of the data according to the mixture */
	/* Another E-step */
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    {
	      sum=0.0;
	      for( b=0; b<2; b++ )
		{
		  if( t==0 )
		    mix_comp_prob[b][n][t]=subdist[i]->state_prob[b]*
		      subdist[i]->subdist[b]->prob( data->sequence[n]->entry[t], NULL );
		  else
		    mix_comp_prob[b][n][t]=subdist[i]->state_prob[b]*
		      subdist[i]->subdist[b]->prob( data->sequence[n]->entry[t], 
						    data->sequence[n]->entry[t-1] );
		  sum+=mix_comp_prob[b][n][t];
		}
	      
	      for( b=0; b<2; b++ )
		mix_comp_prob[b][n][t]/=sum;
	    }
	
	/* Passing weights of the data to the mixture components and creating a mixture of potentially
	   divisible components */
	for( b=0; b<2; b++ )
	  {
	    sum=0.0;
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=0; t<data->sequence[n]->seq_length; t++ )
		{	    
		  mix_comp_prob[b][n][t]*=prob_s[n][t];
		  sum+=mix_comp_prob[b][n][t];
		}
	    
	    temp_dist=subdist[i]->subdist[b];
	    subdist[i]->subdist[b]=new Distribution( DIST_NESTEDDIVIDE, num_states, num_nodes[i] );
	    for( j=0; j<num_nodes[i]; j++ )
	      subdist[i]->subdist[b]->dim_index[j]=dim_index[node_list[i][j]];
	    
	    /* Updating mixture components */
	    subdist[i]->subdist[b]->UpdateNestedParameters( data, mix_comp_prob[b], sum, temp_dist );
	  }
	
	/* Deallocating the weights */
	for( b=0; b<2; b++ )
	  {
	    for( n=0; n<data->num_seqs; n++ )
	      delete mix_comp_prob[b][n];
	    delete mix_comp_prob[b];
	  }
	delete mix_comp_prob;
	
      } /* Creating a mixture of distributions */
  
  
  /* Deallocating node counts */
  for( i=0; i<num_dists+1; i++ )
    delete node_list[i];
  delete node_list;
  delete num_nodes;
  
  delete CLtree;

  return;
}

